using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;

namespace WPF_Util
{
	/// <summary>LabelのTargetプロパティ設定ヘルパーです。</summary>
	public static class LabelTarget
	{

		/// <summary>TargetName添付プロパティを取得します。</summary>
		public static string GetName(Label obj)
		{
			return (string)obj.GetValue(NameProperty);
		}

		/// <summary>TargetName添付プロパティを設定します。</summary>
		public static void SetName(Label obj, string value)
		{
			obj.SetValue(NameProperty, value);
		}

		/// <summary>TargetName添付プロパティを識別します。</summary>
		/// <remarks>
		/// このプロパティはLabel.Targetプロパティの設定ヘルパーです。
		/// XAMLにおける
		/// <code>
		/// &lt;Label Target="{Binding ElementName=target_element_name, Mode=OneTime}" >
		/// </code>
		/// を、下記のように記述できます。
		/// <code>
		/// &lt;Label LabelHelper.TargetName="target_element_name" >
		/// </code>
		/// </remarks>
		public static readonly DependencyProperty NameProperty =
					DependencyProperty.RegisterAttached(
				"Name"
				, typeof(string)
				, typeof(LabelTarget)
				, new UIPropertyMetadata(null, TargetNameChanged)
			);

		/// <summary>TargetNameの変更イベント処理</summary>
		private static void TargetNameChanged(DependencyObject s, DependencyPropertyChangedEventArgs e)
		{
			var label = (Label)s;

			if (e.OldValue != null)
			{
				var bind = BindingOperations.GetBinding(s, Label.TargetProperty);
				if (bind != null && bind.ElementName == (string)e.OldValue)
				{
					BindingOperations.ClearBinding(s, Label.TargetProperty);
				}
			}

			if (e.NewValue != null)
			{

				BindingOperations.SetBinding(s, Label.TargetProperty
					, new Binding() {
						ElementName = e.NewValue.ToString(),
						Mode = BindingMode.OneTime,
					}
				);
			}
		}



		/// <summary>IsTargetNext添付プロパティを取得します。</summary>
		public static bool GetIsNext(Label obj)
		{
			return (bool)obj.GetValue(IsNextProperty);
		}

		/// <summary>IsTargetNext添付プロパティを設定します。</summary>
		public static void SetIsNext(Label obj, bool value)
		{
			obj.SetValue(IsNextProperty, value);
		}

		/// <summary>IsTargetNext添付プロパティを識別します。</summary>
		/// <remarks><para>
		/// このプロパティをtrueに設定したLabelコントロールのTargetプロパティは、
		/// 論理ツリー上の次の要素になります。
		/// </para><para>
		/// 論理ツリーの次の要素がFocasableでない場合はコンテナであると見なして、
		/// その要素のサブツリーを検索して最初のFocasable要素をTargetにします。
		/// </para></remarks>
		public static readonly DependencyProperty IsNextProperty =
					DependencyProperty.RegisterAttached(
				"IsNext"
				, typeof(bool)
				, typeof(LabelTarget)
				, new UIPropertyMetadata(false, IsTargetNextChanged)
			);

		/// <summary>IsTargetNext変更イベント処理</summary>
		private static void IsTargetNextChanged(DependencyObject s, DependencyPropertyChangedEventArgs e)
		{
			var label = (Label)s;

			if ((bool)e.OldValue)
			{
				label.Loaded -= label_Loaded;
			}

			if ((bool)e.NewValue)
			{
				// Loadイベントが起こってたら即座に反映、起こってなければイベント内で処理
				if (label.IsLoaded)
				{
					SetTargetNext(label);
				}
				else
				{
					label.Loaded += label_Loaded;
				}
			}
		}

		/// <summary>LabelのLoadイベントでTargetを設定する</summary>
		static void label_Loaded(object sender, RoutedEventArgs e)
		{
			SetTargetNext((Label)sender);
		}

		/// <summary>論理ツリーでLabelの次の要素をTargetに指定する。</summary>
		private static void SetTargetNext(Label label)
		{
			var parent = LogicalTreeHelper.GetParent(label);
			if (parent == null)
			{
				throw new InvalidOperationException("Labelが論理ツリーに接続されていません。");
			}

			var next = 
				LogicalTreeHelper.GetChildren(parent)
				.Cast<object>()
				.SkipWhile(item => item != label)
				.Skip(1)
				.FirstOrDefault();

			if (next == null)
			{
				if (System.ComponentModel.DesignerProperties.GetIsInDesignMode(label))
				{
					return;
				}
				throw new InvalidOperationException("Labelが論理ツリー兄弟ノードの最後の要素です。");
			}

			var target = (
				from item in GetNodes(next)
				let element = item as UIElement
				where element != null && element.Focusable
				select element
			).FirstOrDefault();

			if (target == null)
			{
				if (System.ComponentModel.DesignerProperties.GetIsInDesignMode(label))
				{
					return;
				}
				throw new InvalidOperationException(
					String.Format(
						"Label[{0}]のTargetになるフォーカス設定可能な要素が見つかりませんでした。"
						, label.Content
					)
				);
			}
			label.Target = target;
		}

		/// <summary>論理ツリーのobj以下の要素をすべて列挙する。</summary>
		private static IEnumerable<object> GetNodes(object obj)
		{
			yield return obj;

			var e = obj as FrameworkElement;
			if (e != null)
			{
				foreach (var child in LogicalTreeHelper.GetChildren(e))
				{
					foreach (var node in GetNodes(child))
					{
						yield return node;
					}
				}
			}

			var ce = obj as FrameworkContentElement;
			if (ce != null)
			{
				foreach (var child in LogicalTreeHelper.GetChildren(ce))
				{
					foreach (var node in GetNodes(child))
					{
						yield return node;
					}
				}
			}
		}


	}
}
